/** @file
 *
 * This library contains constants and functionality for RPC-callable functions,
 *  including DP, query, and model information. Also uncludes functionality for
 *  model initialization and synchronization.
 *
 * @author  Victor Rodrigues (CERN, BE-ICS-SDS)
 */

#uses "fwRPC/fwRPC_common.ctl"

/**
 * RPC function DPEs
 * @ingroup Constants
 * @{
 */
const string fwRPC_FUNCTION_DPT           = "_FwRPC_Function";
const string fwRPC_FUNCTION_DPE_NAME      = ".name"    + fwRPC_DPE_VALUE;
const string fwRPC_FUNCTION_DPE_STUB      = ".stub"    + fwRPC_DPE_VALUE;
const string fwRPC_FUNCTION_DPE_ENABLED   = ".enabled" + fwRPC_DPE_VALUE;
const string fwRPC_FUNCTION_DPE_EXCLUSIVE = ".exclusiveExecution" + fwRPC_DPE_VALUE;
const string fwRPC_FUNCTION_DPE_LIBFILE   = ".libfile" + fwRPC_DPE_VALUE;
const string fwRPC_FUNCTION_DPE_PARAMTYPE = ".paramType" + fwRPC_DPE_VALUE;
const string fwRPC_FUNCTION_DPE_TIMEOUT   = ".timeout_s" + fwRPC_DPE_VALUE;
/* @} */


/**
 * Function model indexes
 * @ingroup Constants
 * @{
 */
const int fwRPC_FUNCTION_MODEL_DP        = 1;
const int fwRPC_FUNCTION_MODEL_NAME      = 2;
const int fwRPC_FUNCTION_MODEL_STUB      = 3;
const int fwRPC_FUNCTION_MODEL_EXCLUSIVE = 4;
const int fwRPC_FUNCTION_MODEL_LIBFILE   = 5;
const int fwRPC_FUNCTION_MODEL_PARAMTYPE = 6;
const int fwRPC_FUNCTION_MODEL_TIMEOUT   = 7;
const int fwRPC_FUNCTION_MODEL_MAXIDX    = 8;
/* @} */


/**
 * Function model misc constants
 * @ingroup Constants
 * @{
 */
const int fwRPC_FUNCTION_MODEL_NO_TIMEOUT = -1;
/* @} */


/**
 * Function query string to retrieve all info.
 * @ingroup Constants
 */
const string fwRPC_FUNCTION_QUERY_UPDATE = "SELECT '" +
  fwRPC_FUNCTION_DPE_NAME      + "', '" +
  fwRPC_FUNCTION_DPE_STUB      + "', '" +
  fwRPC_FUNCTION_DPE_LIBFILE   + "', '" +
  fwRPC_FUNCTION_DPE_PARAMTYPE + "', '" +
  fwRPC_FUNCTION_DPE_EXCLUSIVE + "', '" +
  fwRPC_FUNCTION_DPE_TIMEOUT   + "' " +
  "FROM '*' WHERE _DPT = \"" + fwRPC_FUNCTION_DPT + "\" AND '" +
  fwRPC_FUNCTION_DPE_ENABLED + "' == 1";


//------------------------------------------------------------------------------


/**
 * Construct a new function object.
 *
 * A list of valid parameters may be specified by passing a dyn_int of valid
 * types. This is ANYTYPE_VAR by default, which allows any parameter types to be
 * passed.
 *
 * A timeout may be specified, which limits the time in seconds which the
 * registered procedure may be running before it is terminated.
 *
 * @param funcName  (IN, string)  function name
 * @param stubName  (IN, string)  function stub name
 * @param libFile   (IN, string)  lib file containing the stub function
 * @param paramType (IN, dyn_int) valid param types
 * @param exclusive (IN, bool)    whether the function requires exclusive execution
 * @param timeout   (IN, int)     timeout to apply for procedure execution
 *
 * @returns function object
 *
 * @throws nothing
 */
dyn_mixed fwRPC_functionModel_new(
  string funcName,
  string stubName = "",
  string libFile = "",
  dyn_int paramType = makeDynInt(ANYTYPE_VAR),
  bool exclusive = false,
  int timeout = -1
) {
  dyn_mixed func;

  string dp_function = fwRPC_common_makeDpName(fwRPC_FUNCTION_DPT, funcName);

  func[fwRPC_FUNCTION_MODEL_DP]        = dp_function;
  func[fwRPC_FUNCTION_MODEL_NAME]      = funcName;
  func[fwRPC_FUNCTION_MODEL_STUB]      = stubName;
  func[fwRPC_FUNCTION_MODEL_LIBFILE]   = libFile;
  func[fwRPC_FUNCTION_MODEL_EXCLUSIVE] = exclusive;
  func[fwRPC_FUNCTION_MODEL_PARAMTYPE] = paramType;
  func[fwRPC_FUNCTION_MODEL_TIMEOUT]   = timeout;
  func[fwRPC_FUNCTION_MODEL_MAXIDX]    = 0;

  return func;
}


/**
 * Construct a new function object from a function query result.
 *
 * @param result (IN, dyn_mixed) function query result
 *
 * @returns function object
 *
 * @throws nothing
 */
dyn_mixed fwRPC_functionModel_newFromQuery(dyn_mixed result) {
  dyn_int paramtypes = _fwRPC_functionModel_unpackIntList(result[5]);

  return fwRPC_functionModel_new(
    result[2],
    result[3],
    result[4],
    paramtypes,
    result[6],
    result[7]
  );
}


/**
 * Sync a function object to the current actual state. (pull)
 *
 * @param func (IN|OUT, dyn_mixed) function object to update
 *
 * @returns nothing
 *
 * @throws if func is not a valid function object
 *
 * @reviewed 2018-06-22 @whitelisted{Used in Unit Tests}
 */
void fwRPC_functionModel_sync(dyn_mixed &func) {
  fwRPC_functionModel_checkValid(func);

  dyn_mixed _func = func;
  string dp_func = func[fwRPC_FUNCTION_MODEL_DP];

  string validTypes;

  int err = dpGet(
    dp_func + fwRPC_FUNCTION_DPE_STUB,      _func[fwRPC_FUNCTION_MODEL_STUB],
    dp_func + fwRPC_FUNCTION_DPE_LIBFILE,   _func[fwRPC_FUNCTION_MODEL_LIBFILE],
    dp_func + fwRPC_FUNCTION_DPE_EXCLUSIVE, _func[fwRPC_FUNCTION_MODEL_EXCLUSIVE],
    dp_func + fwRPC_FUNCTION_DPE_PARAMTYPE, validTypes,
    dp_func + fwRPC_FUNCTION_DPE_TIMEOUT,   _func[fwRPC_FUNCTION_MODEL_TIMEOUT]
  );

  dyn_int t = _fwRPC_functionModel_unpackIntList(validTypes);

  _func[fwRPC_FUNCTION_MODEL_PARAMTYPE] = t;

  if (err) {
    fwException_throw(fwRPC_EX_FUNCTION_SYNC, "Error syncing function from DP");
  }

  func = _func;
}


/**
 * Update actual function to a given state. (push)
 *
 * @param func (IN, dyn_mixed) function to update
 *
 * @returns nothing
 *
 * @throws if func is not a valid function object
 */
void fwRPC_functionModel_store(dyn_mixed func) {
  fwRPC_functionModel_checkValid(func);

  string dp_func = func[fwRPC_FUNCTION_MODEL_DP];

  if (!dpExists(dp_func)) {
    dpCreate(dp_func, fwRPC_FUNCTION_DPT);
  }

  string types = _fwRPC_functionModel_packIntList(
    func[fwRPC_FUNCTION_MODEL_PARAMTYPE]
  );

  dpSetWait(
    makeDynString(
      dp_func + fwRPC_FUNCTION_DPE_NAME,
      dp_func + fwRPC_FUNCTION_DPE_STUB,
      dp_func + fwRPC_FUNCTION_DPE_ENABLED,
      dp_func + fwRPC_FUNCTION_DPE_EXCLUSIVE,
      dp_func + fwRPC_FUNCTION_DPE_LIBFILE,
      dp_func + fwRPC_FUNCTION_DPE_PARAMTYPE,
      dp_func + fwRPC_FUNCTION_DPE_TIMEOUT
    ),
    makeDynString(
      func[fwRPC_FUNCTION_MODEL_NAME],
      func[fwRPC_FUNCTION_MODEL_STUB],
      true,
      func[fwRPC_FUNCTION_MODEL_EXCLUSIVE],
      func[fwRPC_FUNCTION_MODEL_LIBFILE],
      types,
      func[fwRPC_FUNCTION_MODEL_TIMEOUT]
    )
  );
}


/**
 * Check if an object is a valid function object, and throw if not.
 *
 * @param func (IN, dyn_mixed) object to test
 *
 * @returns nothing
 *
 * @throws if the given object is not a valid function object
 */
void fwRPC_functionModel_checkValid(dyn_mixed func) {
  if (
    dynlen(func) < fwRPC_FUNCTION_MODEL_MAXIDX ||
    func[fwRPC_FUNCTION_MODEL_DP] == ""
  ) {
    fwException_throw(fwRPC_EX_INVALID_PARAM, "Function is not valid");
  }
}


/**
 *  Compute the timeout date for a given function.
 *  If there is no timeout set, the -1 will be returned.
 *
 *  @param func (IN, dyn_mixed) function object
 *
 *  @returns time value of timeout
 *           -1   if no timeout set
 *
 *  @throws if func is not a valid function object
 */
anytype fwRPC_functionModel_computeTimeout(dyn_mixed func) {
  fwRPC_functionModel_checkValid(func);

  int func_timeout_s = func[fwRPC_FUNCTION_MODEL_TIMEOUT];
  anytype func_timeout;
  if (func_timeout_s <= 0) func_timeout = -1;
  else func_timeout = (time) getCurrentTime() + func_timeout_s;

  return func_timeout;
}


/**
 * @internal
 *
 * Unpacks a semicolon delimited string into an int string. Relies on implicit
 * type cast.
 *
 * @param list (IN, string) delimited string
 *
 * @returns split dyn int
 *
 * @throws nothing
 */
private dyn_int _fwRPC_functionModel_unpackIntList(string list) {
  //ew
  return strsplit(list, ";");
}


/**
 * @internal
 *
 * Packs a dyn_int into a semicolon delimited string. Relies on implicit type
 * cast.
 *
 * @param list (IN, dyn_int) list of ints to pack
 *
 * @returns delimited string
 *
 * @throws nothing
 */
private string _fwRPC_functionModel_packIntList(dyn_int list) {
  string str = list[1];
  int intsToPack = dynlen(list);

  for(int i = 2; i <= intsToPack; i++) {
    str += ";" + (string)list[i];
  }

  return str;
}
