/** @file
 *
 * This library contains the set of functions that handles worker management.
 *
 * Part of the server-side RPC library.
 *
 * @author  Victor Rodrigues (CERN, BE-ICS-SDS)
 */

#uses "fwRPC/fwRPC_common.ctl"
#uses "fwRPC/fwRPC_pmon.ctl"
#uses "fwRPC/fwRPC_workerModel.ctl"
#uses "fwRPC/fwRPC_exception.ctl"
#uses "fwRPC/fwRPC_workerManager.ctl"
#uses "fwRPC/fwRPC_debug.ctl"

/**
 * Mapping of workers.
 * Structure:
 *  key   = worker id
 *  value = worker object (see fwRPC_workerModel.ctl)
 */
private global mapping g_fwRPC_workerManager_workers;

/**
 * Worker iterator for round-robin load balancing
 * value is 0 if no workers exist.
 */
private global int g_fwRPC_workerManager_workersIterator;

/**
 * Callback function to call when a worker is declared dead
 */
private global string g_fwRPC_workerManager_workerDeathCB;


//------------------------------------------------------------------------------


/**
 * Ensures that the system starts with workerCount workers, and is bound to
 *  maxWorkers, unless workerCount is -1, in which case no worker ctrl manager
 *  management is performed (and maxWorkers is therefore ignored).
 * Starts any stopped workers.
 *
 * @param workerCount    (IN, int)    starting number of workers
 * @param maxWorkers     (IN, int)    max number of workers allowed
 * @param workerDeathCB  (IN, string) callback to be executed on worker update event
 *
 * @returns nothing
 *
 * @throws on PMON error
 */
void fwRPC_workerManager_initialize(
  int workerCount=-1, 
  int maxWorkers=-1, 
  string workerDeathCB
) {
  g_fwRPC_workerManager_workersIterator = 0;
  g_fwRPC_workerManager_workerDeathCB = workerDeathCB;
  mappingClear(g_fwRPC_workerManager_workers);

  if (workerCount > -1) {
    if (workerCount > maxWorkers) {
      fwException_throw(fwRPC_EX_INVALID_PARAM,
        "Initial worker count cannot be greater than max workers.");
    }

    _fwRPC_workerManager_balanceWorkerInstances(
      workerCount,
      maxWorkers,
      workerCount
    );
  }
}


/**
 * Notify a heartbeat event for a given worker by it's ID. This function should
 * be called whenever a heartbeat message is received from the worker itself.
 *
 * @param id (IN, int) worker id
 *
 * @returns  0 on success
 *          -1 if the worker doesn't exist
 *
 * @throws nothing
 */
int fwRPC_workerManager_notifyHeartbeat(int id)
synchronized(g_fwRPC_workerManager_workers) {
  if (!mappingHasKey(g_fwRPC_workerManager_workers, id)) return -1;
  g_fwRPC_workerManager_workers[id][fwRPC_WORKER_MODEL_LASTHB] = getCurrentTime();

  return 0;
}


/**
 * Attempt to find a worker which is able to execute a function under the given
 * exclusivity option.
 *
 * @param   exclusiveExec (IN, bool)     whether the request requires exclusive
 *                                       execution
 * @param   worker        (OUT, worker)  a suitable worker, if found
 *
 * @returns  0 on success
 *          -1 if no suitable worker is available
 *
 * @throws  nothing
 */
int fwRPC_workerManager_findSuitableWorker(
  bool exclusiveExec,
  dyn_mixed &worker
) synchronized(g_fwRPC_workerManager_workers) {
  if (mappinglen(g_fwRPC_workerManager_workers) == 0) return -1;

  int workerKey = _fwRPC_workerManager_workersIterator_nextKey();
  dyn_mixed _worker = g_fwRPC_workerManager_workers[workerKey];
  int startingKey = workerKey;

  while(!fwRPC_workerModel_canProcess(_worker, exclusiveExec)) {
    workerKey = _fwRPC_workerManager_workersIterator_nextKey();
    if (workerKey == startingKey) {
      return -1;
    }
    _worker = g_fwRPC_workerManager_workers[workerKey];
  }

  worker = _worker;
  return 0;
}


/**
 * Add a worker to the worker collection if it's not already present. If the
 * worker already exists in the collection then it will not be replaced, and
 * this function will return -1.
 *
 * @param worker   (IN, dyn_mixed) worker to register
 *
 * @returns  0 on success
 *          -1 otherwise
 *
 * @throws nothing
 */
int fwRPC_workerManager_registerWorker(dyn_mixed worker)
synchronized(g_fwRPC_workerManager_workers) {
  int key = worker[fwRPC_WORKER_MODEL_ID];
  if (mappingHasKey(g_fwRPC_workerManager_workers, key)) return -1;

  g_fwRPC_workerManager_workers[key] = worker;
  int tid = startThread("_fwRPC_workerManager_heartmonitor", worker);
  if (tid < 0) {
    fwRPC_dbg_log(
      fwRPC_DBG_EVENT_ERROR,
      __FUNCTION__ + "could not start heartbeat watch for worker " + key
    );
    return -1;
  }

  //reset the worker iterator idx to the first pos if we had no workers before
  if (g_fwRPC_workerManager_workersIterator == 0) {
    g_fwRPC_workerManager_workersIterator = 1;
  }

  fwRPC_dbg_log(
    fwRPC_DBG_EVENT_GENERIC,
    "Registered new worker: " + key
  );

  return 0;
}


/**
 * Delete a worker from the worker collection and it's DP.
 *
 * @param workerId (IN, int) worker to delete
 *
 * @return  0 on success
 *         -1 otherwise
 *
 * @throws nothing
 */
int fwRPC_workerManager_deleteWorker(int workerId)
synchronized(g_fwRPC_workerManager_workers) {
  if (!mappingHasKey(g_fwRPC_workerManager_workers, workerId)) return -1;

  mappingRemove(g_fwRPC_workerManager_workers, workerId);


  //build comm dp, eg: FwRPC_RBMBX_DPT/loc_work_1
  string clientName = fwRPC_DIST_TAG_LOCAL +
                      "_" + fwRPC_DIST_TAG_WORKER +
                      "_" + workerId;

  string rbMbxDp = fwRPC_common_makeDpName(
    fwRPC_RBMBX_DPT,
    clientName
  );

  if (dpExists(rbMbxDp)) dpDelete(rbMbxDp);

  if (mappinglen(g_fwRPC_workerManager_workers) > 0) {
    g_fwRPC_workerManager_workersIterator = 1;
  } else {
    g_fwRPC_workerManager_workersIterator = 0;
  }

  return 0;
}


/**
 * Update a specific worker with a new state.
 *
 * Only status, job count and locked status will be changed.
 *
 * If the worker does not currently exist, it will be added to the pool as an
 * active worker.
 *
 * @param worker (IN, dyn_mixed) New worker state
 *
 * @returns nothing
 *
 * @throws nothing
 */
void fwRPC_workerManager_updateState(dyn_mixed worker)
synchronized(g_fwRPC_workerManager_workers) {
  int id = worker[fwRPC_WORKER_MODEL_ID];

  if (!mappingHasKey(g_fwRPC_workerManager_workers, id)) {
    fwRPC_workerManager_registerWorker(worker);
  } else {
    dyn_mixed local_worker = g_fwRPC_workerManager_workers[id];
    local_worker[fwRPC_WORKER_MODEL_STATUS] = worker[fwRPC_WORKER_MODEL_STATUS];
    local_worker[fwRPC_WORKER_MODEL_JOBS]   = worker[fwRPC_WORKER_MODEL_JOBS];
    local_worker[fwRPC_WORKER_MODEL_LOCKED] = worker[fwRPC_WORKER_MODEL_LOCKED];
    g_fwRPC_workerManager_workers[id] = local_worker;
  }
}


/**
 * Return the total number of registered workers in the local system.
 *
 * @returns number of workers
 *
 * @throws nothing
 */
int fwRPC_workerManager_getWorkerCount() {
  return mappinglen(g_fwRPC_workerManager_workers);
}


/**
 * Set all workers to a pending restart state.
 *
 * The worker manager will attempt to restart the workers immediately. If this
 * is done, then the function will return false to indicate that the worker
 * manager is not pending restart. If this is not the case then the function
 * will return true, indicating that the worker manager is still pending restart
 * on at least one worker.
 *
 * @param immediateRestart (IN, bool) whether we can try to restart immediately
 *                                    if a worker is idle
 *
 * @returns true if the worker manager is pending restart
 *
 * @throws nothing
 */
bool fwRPC_workerManager_pendWorkerRestart(bool immediateRestart)
synchronized (g_fwRPC_workerManager_workers) {
  dyn_anytype keys = mappingKeys(g_fwRPC_workerManager_workers);

  bool pendingRestart = false;

  for (int i=1; i<=dynlen(keys); ++i) {
    dyn_mixed worker = g_fwRPC_workerManager_workers[keys[i]];
    if (immediateRestart && fwRPC_workerModel_isIdle(worker)) {
      fwRPC_workerManager_restartWorker(worker);
    } else {
      g_fwRPC_workerManager_workers[keys[i]][fwRPC_WORKER_MODEL_PEND_RESTART] = true;
      pendingRestart = true;
    }
  }

  return pendingRestart;
}


/**
 * Return whether or not a given worker is pending restart.
 *
 * @param worker (IN, dyn_mixed) worker object to check
 *
 * @returns true  if worker is pending restart
 *          false otherwise
 *
 * @throws nothing
 */
bool fwRPC_workerManager_isWorkerPendingRestart(dyn_mixed worker) {
  int workerId = worker[fwRPC_WORKER_MODEL_ID];

  if (!mappingHasKey(g_fwRPC_workerManager_workers, workerId)) return false;

  return g_fwRPC_workerManager_workers[workerId][fwRPC_WORKER_MODEL_PEND_RESTART];
}


/**
 * Return whether or not any workers are pending restart.
 *
 * @returns true  if at least one worker is pending restart
 *          false if no workers are pending restart
 *
 * @throws nothing
 */
bool fwRPC_workerManager_isAnyWorkerPendingRestart() {
  dyn_anytype keys = mappingKeys(g_fwRPC_workerManager_workers);
  for (int i=1; i<=dynlen(keys); ++i) {
    if (g_fwRPC_workerManager_workers[keys[i]][fwRPC_WORKER_MODEL_PEND_RESTART]) {
      return true;
    }
  }

  return false;
}


/**
 * Restart a given worker.
 *
 * @param worker (IN, dyn_mixed) worker to restart
 *
 * @returns nothing
 *
 * @throws nothing
 */
void fwRPC_workerManager_restartWorker(dyn_mixed worker)
synchronized(g_fwRPC_workerManager_workers) {
  int manId = worker[fwRPC_WORKER_MODEL_ID];

  g_fwRPC_workerManager_workers[manId][fwRPC_WORKER_MODEL_STATUS] = fwRPC_WORKER_STATUS_RESTARTING;

  dyn_mixed conInfo = fwRPC_pmon_getConInfo();

  fwRPC_dbg_log(
    fwRPC_DBG_EVENT_GENERIC,
    "Restarting worker: " + manId
  );

  g_fwRPC_workerManager_workers[manId][fwRPC_WORKER_MODEL_PEND_RESTART] = false;

  int index = -1;
  dyn_dyn_mixed managers;
  fwRPC_pmon_getWorkerCtrlManagers(conInfo, managers);

  int managersToCheck = dynlen(managers);
  for (int i=managersToCheck; i > 0; --i) {
    dyn_mixed mngr = managers[i];
    if (mngr[fwRPC_PMON_MANINFO_CMD] != fwRPC_WORKER_SCRIPT) continue;
    if (mngr[fwRPC_PMON_MANINFO_NUM] != manId) continue;

    index = mngr[fwRPC_PMON_MANINFO_INDEX];
    break;
  }

  if (index == -1) {
    fwRPC_dbg_log(
      fwRPC_DBG_EVENT_ERROR,
      __FUNCTION__ + ": could not restart worker, mngr not found for worker " +
      manId
    );
    return;
  }

  fwRPC_pmon_stopCtrlManager(conInfo, index);
  fwRPC_pmon_startCtrlManager(conInfo, index);
}


//------------------------------------------------------------------------------


/**
 * @internal
 *
 * Balance worker instances
 *
 * @param min      (IN, int) minimum concurrent worker count
 * @param max      (IN, int) maximum concurrent worker count
 * @param starting (IN, int) initial worker count
 *
 * @returns nothing
 *
 * @throws nothing
 *
 * @pre min <= starting <= max
 */
private void _fwRPC_workerManager_balanceWorkerInstances(
  int min,
  int max,
  int starting
) {
  dyn_mixed conInfo;
  dyn_dyn_mixed workers;
  int numWorkers;
  dyn_int inactiveWorkers;
  int activeWorkers = 0;

  conInfo = fwRPC_pmon_getConInfo();
  fwRPC_pmon_getWorkerCtrlManagers(conInfo, workers);

  //ensure current #workers >= min && #workers <= max
  numWorkers = dynlen(workers);
  if (numWorkers < min) {
    int workersToAppend = min-numWorkers;
    for (int i=0; i<workersToAppend; i++) {
      fwRPC_pmon_appendCtrlManager(conInfo, "fwRPC/fwRPC_worker_script.ctl");
    }
  } else if (numWorkers > max) {
    int workersToDelete = numWorkers-max;
    for (int i=0; i<workersToDelete; i++) {
      fwRPC_pmon_deleteLastWorker(conInfo);
    }
  }

  //Find inactive workers while active count < starting param
  fwRPC_pmon_getWorkerCtrlManagers(conInfo, workers);
  numWorkers = dynlen(workers);
  for (int i=1; (i <= numWorkers) && (activeWorkers < starting); i++) {
    dyn_mixed worker = workers[i];
    int state = worker[fwRPC_PMON_MANINFO_STATE];

    switch(state) {
    case fwRPC_PMON_MANINFO_STATE_INIT:
    case fwRPC_PMON_MANINFO_STATE_RUNNING:
      activeWorkers++;
      break;

    case fwRPC_PMON_MANINFO_STATE_STOPPED:
      dynAppend(inactiveWorkers, i);
      break;

    case fwRPC_PMON_MANINFO_STATE_BLOCKED:
      fwRPC_dbg_log(
        fwRPC_DBG_EVENT_ERROR,
        __FUNCTION__ + ": worker is blocked: " + worker[fwRPC_WORKER_MODEL_ID]
      );
      break;

    default:
      fwRPC_dbg_log(
        fwRPC_DBG_EVENT_ERROR,
        __FUNCTION__ + ": Unknown error state for worker: " +
        worker[fwRPC_WORKER_MODEL_ID] + " -- " + state
      );
    }
  }

  //activate inactive workers if necessary
  numWorkers = dynlen(inactiveWorkers);
  for (int i=1; i<=numWorkers && activeWorkers < starting; ++i) {
    try {
      fwRPC_pmon_startCtrlManager(
        conInfo,
        workers[inactiveWorkers[i]][fwRPC_PMON_MANINFO_INDEX]
      );
      activeWorkers++;
    } catch {
      fwRPC_dbg_log(
        fwRPC_DBG_EVENT_ERROR,
        "Error starting WCCOActrl manager: " +
        worker[inactiveWorkers[i]][fwRPC_PMON_MANINFO_INDEX] +
        " -- " + fwException_get() //TODO: does this print well?
      );
    }
  }
}


/**
 * @internal
 *
 * 
 * Handle death event of a worker. Calls death event CB specified during
 * intialisation of the worker manager.
 *
 * @param workerId (IN, int) id of dead worker
 *
 * @returns nothing
 *
 * @throws nothing
 */
private void _fwRPC_workerManager_workerPostMortem(int workerId) {
  int tid = startThread(g_fwRPC_workerManager_workerDeathCB, workerId);
  if (tid == -1) {
    fwRPC_dbg_log(
      fwRPC_DBG_EVENT_ERROR,
      __FUNCTION__ + ": Could not call death event function: " +
      g_fwRPC_workerManager_workerDeathCB
    );
  }
}


/**
 * @internal
 *
 * Monitor the heartbeat of a given worker. Executes post mortem function in the
 *  event of death.
 *
 * @param worker (IN, dyn_mixed) worker to monitor
 *
 * @returns nothing
 *
 * @throws nothing
 *
 * @reviewed 2018-06-21 @whitelisted{Thread}
 */
private void _fwRPC_workerManager_heartmonitor(dyn_mixed worker) {
  bool current;
  bool expired;
  dyn_anytype response;
  int workerId = worker[fwRPC_WORKER_MODEL_ID];

  //wait a bit for the worker to fully initialize before detecting for heartbeat
  if (!fwRPC_pmon_waitForWorkerStart(workerId)) {
    _fwRPC_workerManager_workerPostMortem(workerId);
    fwRPC_dbg_log(
      fwRPC_DBG_EVENT_ERROR,
      __FUNCTION__ + "Could not wait for worker start: " + workerId
    );
    return;
  }

  while(true) {
    delay(
      (fwRPC_WORKER_HEARTBEAT_RATE_MS / 1000) % 60,
      fwRPC_WORKER_HEARTBEAT_RATE_MS % 1000
    );

    if (!mappingHasKey(g_fwRPC_workerManager_workers, workerId)) {
      _fwRPC_workerManager_workerPostMortem(workerId);
      return;
    }

    worker = g_fwRPC_workerManager_workers[workerId];
    if (worker[fwRPC_WORKER_MODEL_STATUS] == fwRPC_WORKER_STATUS_RESTARTING) {
      continue;
    }

    time dt = getCurrentTime() - worker[fwRPC_WORKER_MODEL_LASTHB];
    if (second(dt) > fwRPC_WORKER_HEARTBEAT_WAIT_S) {
      fwRPC_dbg_log(
        fwRPC_DBG_EVENT_WARNING,
        __FUNCTION__ + ": worker expired: " + workerId
      );

      _fwRPC_workerManager_workerPostMortem(workerId);
      return;
    }
  }
}


/**
 * @internal
 *
 * Advance the worker iterator. Used to keep index on the previously assigned
 *  worker, to achieve round-robin task distribution. If there are no available
 *  workers, function returns 0.
 *
 * @returns index of the next worker.
 *          0 if no workers are available.
 *
 * @throws nothing
 */
private int _fwRPC_workerManager_workersIterator_nextKey()
synchronized(g_fwRPC_workerManager_workers) {
  int numWorkers = mappinglen(g_fwRPC_workerManager_workers);

  if (!numWorkers) return 0;

  g_fwRPC_workerManager_workersIterator = ++g_fwRPC_workerManager_workersIterator % numWorkers + 1;
  if (g_fwRPC_workerManager_workersIterator < numWorkers) {
    g_fwRPC_workerManager_workersIterator++;
  } else {
    g_fwRPC_workerManager_workersIterator = 1; 
  }

  return mappingGetKey(
    g_fwRPC_workerManager_workers,
    g_fwRPC_workerManager_workersIterator
  );
}
