/**@file 

// unDeviceTypeCreatorDeviceFunctions.ctl
This library contains functions used to manipulate device types in the package creation application.

@par Creation Date
  30/11/2011

@par Modification History
  30/11/2011 - First release
  
@par Constraints
  Can only be used by the main panel of the package creation application (dependencies to variables and widgets of the panel)

@par Usage
  Public

@par PVSS managers
  Ui, CTRL

@author 
  Cyril Caillaba (EN-ICE)
*/


#uses "libunCore/unDeviceTypeCreatorDeprecated.ctl"

//@{

// ------------------------------------------------------------------------------------ //
// ------------------unDeviceTypeCreatorDevice_addNewDeviceType--------------------- //
// ------------------------------------------------------------------------------------ //
/** Adds a new device type in the list.
  
@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  Ui, CTRL

@param sDeviceTypeName  input, name of the device type to add to the list.
@param bAlreadyExists input, whether the device already exists somewhere else or not.
@param bBefore intput, true to insert before the currently selected device, false to insert after.
@param dsExceptions output, the list of exceptions that occur in the function.
*/
void unDeviceTypeCreatorDevice_addNewDeviceType(
    string sDeviceTypeName,
    bool bAlreadyExists,
    bool bBefore,
    dyn_string &dsExceptions)
{
  
  // Check if the name is empty
  if (sDeviceTypeName == "")
  {
    string sError = getCatStr("unDeviceTypeCreator","EDIT-EMPTYDEVICENAME");
      
    fwException_raise(
        dsExceptions,
        "ERROR",
        sError,
        "EDIT-EMPTYDEVICENAME"
    ); 
    return; 
  }
  
  // Always replace spaces by underscores
  strreplace(sDeviceTypeName, " ", "_");
  
    
  shape tableDevice = getShape("tableDeviceList"); 
  
  // Check if this device type is not already in the list
  for (int i = 0 ; i < tableDevice.lineCount() ; i++)
  {
    string sCurrentDeviceTypeName;
    getValue("tableDeviceList","cellValueRC",i,COLUMN_DEVICE_NAME,sCurrentDeviceTypeName);  
    if (sCurrentDeviceTypeName == sDeviceTypeName)
    {
      string sError = getCatStr("unDeviceTypeCreator","EDIT-DEVICEALREADYEXISTS");
      strreplace(sError,"DEVICE",sDeviceTypeName);
      
      fwException_raise(
        dsExceptions,
        "ERROR",
        sError,
        "EDIT-DEVICEALREADYEXISTS"
      ); 
      return; 
    }
      
  }    
 
  if (currentCellDevice > -1)
  {
    if (bBefore)
    {
      tableDevice.insertLineN(currentCellDevice);
      tableDevice.cellValueRC(currentCellDevice,COLUMN_DEVICE_NAME,sDeviceTypeName);
      tableDevice.cellValueRC(currentCellDevice,COLUMN_DEVICE_EXISTS,bAlreadyExists);
    }
    else
    {
      tableDevice.insertLineN(currentCellDevice+1);
      tableDevice.cellValueRC(currentCellDevice+1,COLUMN_DEVICE_NAME,sDeviceTypeName);
      tableDevice.cellValueRC(currentCellDevice+1,COLUMN_DEVICE_EXISTS,bAlreadyExists);
    }
    
  }
  else
  {
    if (bBefore)
    {
      tableDevice.insertLineN(0);
      tableDevice.cellValueRC(0,COLUMN_DEVICE_NAME,sDeviceTypeName);
      tableDevice.cellValueRC(0,COLUMN_DEVICE_EXISTS,bAlreadyExists);
    }
    else 
    {
      tableDevice.appendLine(COLUMN_DEVICE_NAME, sDeviceTypeName, COLUMN_DEVICE_EXISTS, bAlreadyExists);
    }
  }

  // Clear the selected line, so that we add at the end for the next device
  currentCellDevice = -1;
  
  // Clear the content of the text field
  textFieldNewDeviceType.text("");
}

// ------------------------------------------------------------------------------------ //
// -------------------unDeviceTypeCreatorDevice_deleteDeviceType-------------------- //
// ------------------------------------------------------------------------------------ //

/** Deletes the currently selected device type in the list.
  
@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  Ui, CTRL
*/
void unDeviceTypeCreatorDevice_deleteDeviceType()
{
  if (currentCellDevice == -1)
  {
    return;
  } 
  else
  {
    setValue("tableDeviceList","deleteLineN",currentCellDevice);
	// After the remove, clear the selection.
    currentCellDevice = -1;
  }
}

// ------------------------------------------------------------------------------------ //
// ----------------unDeviceTypeCreatorDevice_deleteAllDeviceTypes------------------- //
// ------------------------------------------------------------------------------------ //

/** Deletes all the device types in the list.
  
@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  Ui, CTRL
*/
void unDeviceTypeCreatorDevice_deleteAllDeviceTypes()
{
  int lineCount;
  getValue("tableDeviceList","lineCount",lineCount);
  setValue("tableDeviceList","deleteLinesN",0,lineCount);
  currentCellDevice = -1;
}

//@}
