/**@file 

// unDeviceTypeCreatorFrontEndFunctions.ctl
This library contains functions used to manipulate front-end types in the package creation application.

@par Creation Date
  30/11/2011

@par Modification History
  30/11/2011 - First release
  
@par Constraints
  Can only be used by the main panel of the package creation application (dependencies to variables and widgets of the panel)

@par Usage
  Public

@par PVSS managers
  Ui, CTRL

@author 
  Cyril Caillaba (EN-ICE)
*/

#uses "libunCore/unDeviceTypeCreatorDeprecated.ctl"

//@{


// ------------------------------------------------------------------------------------ //
// -----------------unDeviceTypeCreatorFrontEnd_addNewFrontEndType------------------ //
// ------------------------------------------------------------------------------------ //
/** Adds a new front-end type in the list.
  
@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  Ui, CTRL

@param sFrontEndTypeName  input, name of the front-end type to add to the list.
@param bAlreadyExists input, whether the front-end already exists somewhere else or not.
@param bBefore intput, true to insert before the currently selected front-end, false to insert after.
@param dsExceptions output, the list of exceptions that occur in the function.
*/
void unDeviceTypeCreatorFrontEnd_addNewFrontEndType(
    string sFrontEndTypeName,
    bool bAlreadyExists,
    bool bBefore,
	dyn_string &dsExceptions)
{
  
  // Check if the name is empty
  if (sFrontEndTypeName == "")
  {
    string sError = getCatStr("unDeviceTypeCreator","EDIT-EMPTYFRONTENDNAME");
      
    fwException_raise(
        dsExceptions,
        "ERROR",
        sError,
        "EDIT-EMPTYFRONTENDNAME"
    ); 
    return; 
  }
  
  // Always replace spaces by underscores
  strreplace(sFrontEndTypeName, " ", "_");
  
  shape tableFrontEnd = getShape("tableFrontEndList"); 
  
  // Check if this front-end type is not already in the list
  for (int i = 0 ; i < tableFrontEnd.lineCount() ; i++)
  {
    string currentFrontEndTypeName;
    getValue("tableFrontEndList","cellValueRC",i,COLUMN_FRONT_END_NAME,currentFrontEndTypeName);  
    if (currentFrontEndTypeName == sFrontEndTypeName)
    {
      string sError = getCatStr("unDeviceTypeCreator","EDIT-FRONTENDALREADYEXISTS");
      strreplace(sError,"FRONTEND",sFrontEndTypeName);
      
      fwException_raise(
        dsExceptions,
        "ERROR",
        sError,
        "EDIT-FRONTENDALREADYEXISTS"
      ); 
      return; 
    }
      
  }    
 
  if (currentCellFrontEnd > -1)
  {
    if (bBefore)
    {
      tableFrontEnd.insertLineN(currentCellFrontEnd);
      tableFrontEnd.cellValueRC(currentCellFrontEnd,COLUMN_FRONT_END_NAME,sFrontEndTypeName);
      tableFrontEnd.cellValueRC(currentCellFrontEnd,COLUMN_FRONT_END_EXISTS,bAlreadyExists);
    }
    else
    {
      tableFrontEnd.insertLineN(currentCellFrontEnd+1);
      tableFrontEnd.cellValueRC(currentCellFrontEnd+1,COLUMN_FRONT_END_NAME,sFrontEndTypeName);
      tableFrontEnd.cellValueRC(currentCellFrontEnd+1,COLUMN_FRONT_END_EXISTS,bAlreadyExists);
    }
    
  }
  else
  {
    if (bBefore)
    {
      tableFrontEnd.insertLineN(0);
      tableFrontEnd.cellValueRC(0,COLUMN_FRONT_END_NAME,sFrontEndTypeName);
      tableFrontEnd.cellValueRC(0,COLUMN_FRONT_END_EXISTS,bAlreadyExists);
    }
    else 
    {
      tableFrontEnd.appendLine(COLUMN_FRONT_END_NAME, sFrontEndTypeName, COLUMN_FRONT_END_EXISTS, bAlreadyExists);
    }
  }

  // Clear the selected line, so that we add at the end for the next front-end
  currentCellFrontEnd = -1;
  
  // Clear the content of the text field
  textFieldNewFrontEndType.text("");
}

// ------------------------------------------------------------------------------------ //
// ------------------unDeviceTypeCreatorFrontEnd_deleteFrontEndType----------------- //
// ------------------------------------------------------------------------------------ //
/** Deletes the currently selected front-end type in the list.
  
@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  Ui, CTRL
*/
void unDeviceTypeCreatorFrontEnd_deleteFrontEndType()
{
  if (currentCellFrontEnd == -1)
  {
    return;
  } 
  else
  {
    setValue("tableFrontEndList","deleteLineN",currentCellFrontEnd);
    currentCellFrontEnd = -1;
  }
}


// ------------------------------------------------------------------------------------ //
// --------------unDeviceTypeCreatorFrontEnd_deleteAllFrontEndTypes----------------- //
// ------------------------------------------------------------------------------------ //

/** Deletes all the front-end types of the list.
  
@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  Ui, CTRL
*/
void unDeviceTypeCreatorFrontEnd_deleteAllFrontEndTypes()
{
  int lineCount;
  getValue("tableFrontEndList","lineCount",lineCount);
  setValue("tableFrontEndList","deleteLinesN",0,lineCount);
  currentCellFrontEnd = -1;
}

//@}
