/**@file
 
// unDeviceTypeCreatorWidgetFunctions.ctl
This library contains functions used to manipulate widget types in the package creation application.

@par Creation Date
  30/11/2011

@par Modification History
  30/11/2011 - First release
  
@par Constraints
  Can only be used by the main panel of the package creation application (dependencies to variables and widgets of the panel)

@par Usage
  Public

@par PVSS managers
  Ui, CTRL

@author 
  Cyril Caillaba (EN-ICE)
*/

#uses "libunCore/unDeviceTypeCreatorDeprecated.ctl"

//@{
// ------------------------------------------------------------------------------------ //
// ------------------unDeviceTypeCreatorWidget_addNewWidgetType--------------------- //
// ------------------------------------------------------------------------------------ //
/** Adds a new widget type in the list.
  
@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  Ui, CTRL

@param sWidgetTypeName  input, name of the front-end type to add to the list.
@param bAlreadyExists input, whether the front-end already exists somewhere else or not.
@param bBefore intput, true to insert before the currently selected front-end, false to insert after.
@param dsExceptions output, the list of exceptions that occur in the function.
*/
void unDeviceTypeCreatorWidget_addNewWidgetType(
    string sWidgetTypeName,
    bool bAlreadyExists,
    bool bBefore,
	dyn_string &dsExceptions)
{
  
  // Check if the name is empty
  if (sWidgetTypeName == "")
  {
    string sError = getCatStr("unDeviceTypeCreator","EDIT-EMPTYWIDGETNAME");
      
    fwException_raise(
        dsExceptions,
        "ERROR",
        sError,
        "EDIT-EMPTYWIDGETNAME"
    ); 
    return; 
  }
  
  // Always replace spaces by underscores
  strreplace(sWidgetTypeName, " ", "_");
  
    
  shape tableWidget = getShape("tableWidgetList"); 
  
  // Check if this widget type is not already in the list
  for (int i = 0 ; i < tableWidget.lineCount() ; i++)
  {
    string currentWidgetTypeName;
    getValue("tableWidgetList","cellValueRC",i,COLUMN_WIDGET_NAME,currentWidgetTypeName);  
    if (currentWidgetTypeName == sWidgetTypeName)
    {
      string sError = getCatStr("unDeviceTypeCreator","EDIT-WIDGETALREADYEXISTS");
      strreplace(sError,"WIDGET",sWidgetTypeName);
      
      fwException_raise(
        dsExceptions,
        "ERROR",
        sError,
        "EDIT-WIDGETALREADYEXISTS"
      ); 
      return; 
    }
      
  }    
 
  if (currentCellWidget > -1)
  {
    if (bBefore)
    {
      tableWidget.insertLineN(currentCellWidget);
      tableWidget.cellValueRC(currentCellWidget,COLUMN_WIDGET_NAME,sWidgetTypeName);
      tableWidget.cellValueRC(currentCellWidget,COLUMN_WIDGET_EXISTS,bAlreadyExists);
    }
    else
    {
      tableWidget.insertLineN(currentCellWidget+1);
      tableWidget.cellValueRC(currentCellWidget+1,COLUMN_WIDGET_NAME,sWidgetTypeName);
      tableWidget.cellValueRC(currentCellWidget+1,COLUMN_WIDGET_EXISTS,bAlreadyExists);
    }
    
  }
  else
  {
    if (bBefore)
    {
      tableWidget.insertLineN(0);
      tableWidget.cellValueRC(0,COLUMN_WIDGET_NAME,sWidgetTypeName);
      tableWidget.cellValueRC(0,COLUMN_WIDGET_EXISTS,bAlreadyExists);
    }
    else 
    {
      tableWidget.appendLine(COLUMN_WIDGET_NAME, sWidgetTypeName, COLUMN_WIDGET_EXISTS, bAlreadyExists);
    }
  }

  // Clear the selected line, so that we add at the end for the next device
  currentCellWidget = -1;
  
  // Clear the content of the text field
  textFieldNewWidgetType.text("");
}

// ------------------------------------------------------------------------------------ //
// -------------------unDeviceTypeCreatorWidget_deleteWidgetType-------------------- //
// ------------------------------------------------------------------------------------ //
/** Deletes the currently selected widget type in the list.
  
@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  Ui, CTRL
*/
void unDeviceTypeCreatorWidget_deleteWidgetType()
{
  if (currentCellWidget == -1)
  {
    return;
  } 
  else
  {
    setValue("tableWidgetList","deleteLineN",currentCellWidget);
    currentCellWidget = -1;
  }
}


// ------------------------------------------------------------------------------------ //
// ----------------unDeviceTypeCreatorWidget_deleteAllWidgetTypes------------------- //
// ------------------------------------------------------------------------------------ //
/** Deletes all the widget types of the list.
  
@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  Ui, CTRL
*/
void unDeviceTypeCreatorWidget_deleteAllWidgetTypes()
{
  int lineCount;
  getValue("tableWidgetList","lineCount",lineCount);
  setValue("tableWidgetList","deleteLinesN",0,lineCount);
  currentCellWidget = -1;
}

//@}
